/*
 * Copyright 2023, Rainer Noack, Frankfurt am Main, Germany.
 *
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. *
 */
package com.enitsys.sqltools;

import java.math.BigDecimal;
import java.sql.*;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeFormatterBuilder;
import java.util.Locale;
import java.util.Properties;

public enum BaseFormat {
		STRING("%1$s") {
			@Override
			public Object getValue(ResultSet rs, int col) throws SQLException {
				return rs.getString(col);
			}
		},
		NUMBER("%1$f") {
			@Override
			public Object getValue(ResultSet rs, int col) throws SQLException {
				return rs.getBigDecimal(col);
			}
		},
		INTEGER("%1$d") {
			@Override
			public Object getValue(ResultSet rs, int col) throws SQLException {
				BigDecimal bd=rs.getBigDecimal(col);
				return bd==null ? null : bd.unscaledValue();
			}
		},
		DATE("%1$tF") {
			@Override
			public Object getValue(ResultSet rs, int col) throws SQLException {
				return rs.getDate(col).toLocalDate();
			}
			@SuppressWarnings("rawtypes")
			@Override
			public Comparable parse(String text, DateTimeFormatter formatter) {
				return LocalDate.parse(text, formatter);
			}
			@Override
			public String getSQLString(Object value) {
				return value==null?null:((LocalDate)value).format(DateTimeFormatter.ISO_LOCAL_DATE);
			}
		}, 
		DATETIME("%1$tF %1$tT") {
			@Override
			public Object getValue(ResultSet rs, int col) throws SQLException {
				return rs.getTimestamp(col).toLocalDateTime();
			}
			@SuppressWarnings("rawtypes")
			@Override
			public Comparable parse(String text, DateTimeFormatter formatter) {
				return LocalDateTime.parse(text, formatter);
			}
			@Override
			public String getSQLString(Object value) {
				return value==null?null:((LocalDateTime)value).format(DT_ISO);
			}
		}, 
		TIME("%1$tT") {
			@Override
			public Object getValue(ResultSet rs, int col) throws SQLException {
				return rs.getTime(col).toLocalTime();
			}
			@SuppressWarnings("rawtypes")
			@Override
			public Comparable parse(String text, DateTimeFormatter formatter) {
				return LocalTime.parse(text, formatter);
			}
			@Override
			public String getSQLString(Object value) {
				return value==null?null:((LocalTime)value).format(DateTimeFormatter.ISO_LOCAL_TIME);
			}
		},
		TIMESTAMP("%1$tF %1$tT.%1$tN") {
			public Object getValue(ResultSet rs, int col) throws SQLException {
				return rs.getTimestamp(col).toLocalDateTime();
			}
			@SuppressWarnings("rawtypes")
			@Override
			public Comparable parse(String text, DateTimeFormatter formatter) {
				return LocalDateTime.parse(text, formatter);
			}
			@Override
			public String getSQLString(Object value) {
				return value==null?null:((LocalDateTime)value).format(TS_JDBC);
			}
		};
		protected static final DateTimeFormatter DT_ISO = new DateTimeFormatterBuilder().appendPattern("yyyy-MM-dd HH:mm:ss").toFormatter(Locale.US);
		protected static final DateTimeFormatter TS_JDBC = new DateTimeFormatterBuilder().appendPattern("yyyy-MM-dd HH:mm:ss.SSSSSSSSS").toFormatter(Locale.US);
		private final String defaultFormat;
		BaseFormat(String defaultFormat) {
			this.defaultFormat=defaultFormat;
		}
		public String getFormat(Properties taskProperties) { /*
			String x=taskProperties.getProperty("format."+name());
			if(x!=null)
				System.out.println("Property found: format."+name()+"="+x);
				*/
			return taskProperties.getProperty("format."+name(), defaultFormat);
		}
		@SuppressWarnings("rawtypes")
		public Comparable parse(String text, DateTimeFormatter formatter) {
			return null;
		}
		abstract public Object getValue(ResultSet rs, int col) throws SQLException;
		public String getSQLString(Object value) {
			return value==null?null:String.format(Locale.US, defaultFormat, value);
		}
		private static BaseFormat valueOf(int sqltype, int precision, int scale) {
			switch (sqltype) {
			case Types.BIGINT:
			case Types.BIT:
			case Types.BOOLEAN:
			case Types.INTEGER:
			case Types.SMALLINT:
			case Types.TINYINT:
				return BaseFormat.INTEGER;
			case Types.DECIMAL:
			case Types.NUMERIC:
				if (scale == 0 && precision != 0)
					return BaseFormat.INTEGER;
				else
					return BaseFormat.NUMBER;
			case Types.DOUBLE:
			case Types.FLOAT:
			case Types.REAL:
				return BaseFormat.NUMBER;
			case Types.DATE:
				return BaseFormat.DATE;
			case Types.TIME:
				return BaseFormat.TIME;
			case Types.TIMESTAMP:
				return BaseFormat.TIMESTAMP;
			}
			return BaseFormat.STRING;
		}
		public static BaseFormat valueOf(int sqltype, int precision, int scale, String sqltypename, String colname, Properties sqlProperties, Properties taskProperties) {
			return BaseFormat.valueOf(taskProperties.getProperty("treat.column."+colname
					 ,taskProperties.getProperty("treat."+sqltypename
							,sqlProperties.getProperty("treat."+sqltypename
							, BaseFormat.valueOf(sqltype, precision, scale).name()))));
		}
		public String getFormat(Properties taskProperties, String columnname) {
			return taskProperties.getProperty("column."+columnname, getFormat(taskProperties));
		}

	}