/*
 * Copyright 2023, Rainer Noack, Frankfurt am Main, Germany.
 *
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. *
 */
package com.enitsys.sqltools;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.PrintStream;
import java.nio.charset.StandardCharsets;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Properties;
import java.util.regex.Pattern;

import com.enitsys.sqltools.CSVFormat.CSVDefinition;
import com.enitsys.sqltools.CSVFormat.ColumnFormat;
import com.enitsys.sqltools.CSVFormat.StringFormat;


public final class CSVAnalyser extends AbstractTool {
	
	static class ColumnAnalyser {
		private int nulls=0;
		private int notNulls=0;
//		private String header=null;
		ColumnFormat[] qualifiedColumnFormats;
		StringFormat stringFormatDefinition=new StringFormat();
		ColumnAnalyser(Collection<ColumnFormat> formatDefinitions) {
			this.qualifiedColumnFormats=formatDefinitions.toArray(new ColumnFormat[formatDefinitions.size()]);
			for(int i=0;i<this.qualifiedColumnFormats.length;i++)
				this.qualifiedColumnFormats[i]=this.qualifiedColumnFormats[i].clone();
		}
		private boolean checkNull(String value) {
			if (value==null || value.length()==0) {
				nulls++;
				return true;
			}
			notNulls++;
			return false;
		}
		void analyse(String value) {
			if(!checkNull(value)) {
				stringFormatDefinition.matches(value);
				for(int i=0;i<qualifiedColumnFormats.length;i++)
					if(qualifiedColumnFormats[i]!=null &&!qualifiedColumnFormats[i].matches(value))
						qualifiedColumnFormats[i]=null;
			}
		}
		public Collection<ColumnFormat> getFormats() {
			ArrayList<ColumnFormat> formats=new ArrayList<>();
			for(ColumnFormat format:qualifiedColumnFormats)
				if(format!=null)
					formats.add(format);
			return formats;
		}
		public String getHeaderValue() {
			return stringFormatDefinition.constantValue();
		}
		public boolean isAllwaysNull() {
			return notNulls==0; 
		}
		public Boolean isHeader(ColumnAnalyser header) {
			if(header==null)
				return true;
			if(!header.isAllwaysNull()&&!header.getFormats().isEmpty())
				return false;
			if(isAllwaysNull()&&!header.isAllwaysNull())
				return true;
			if(!isAllwaysNull()&&!getFormats().isEmpty())
				return true;
			return null;
		}
		public void write(PrintStream out, String prefix) {
			int i=0;
			Collection<ColumnFormat> formats=getFormats();
			for (ColumnFormat format: formats) {
				if (i==0 || prefix.startsWith("#")) 
					format.write(out, prefix);
				else 
					format.write(out, "## "+prefix);
				i++;
			}
			if (i==0 || prefix.startsWith("#")) 
				stringFormatDefinition.write(out, prefix);
			else 
				stringFormatDefinition.write(out, "## "+prefix);
		}
		public String ddlFormat(Properties ddlProperties) {
			Collection<ColumnFormat> formats=getFormats();
			if (formats.isEmpty())
				return stringFormatDefinition.ddlFormat(ddlProperties);
			for (ColumnFormat format: formats)
				return format.ddlFormat(ddlProperties);
			return null;
		}
	}

	private CSVFormat csvFormat = new CSVFormat();
	private File inputFile;
	private File propertiesOutputFile;
	private File ddlOutputFile;
	private final Case casesensitivity;
	private Properties ddlProperties=new Properties();
	public CSVAnalyser(String taskname, String directory, String inputFilename)throws IOException {
		super( taskname, directory, true);
		csvFormat.loadFormatDefinitions(configDir, taskProperties);
		this.inputFile=inputFile(inputFilename);
		
		propertiesOutputFile=new File(destDir,taskname+".properties").getAbsoluteFile();
		ddlOutputFile=new File(destDir,taskname+".sql").getAbsoluteFile();
		
		String rdbms=System.getProperty("rdbms", "Oracle");
		loadDDLProperties(ddlProperties, rdbms);
		casesensitivity=Case.valueOf(System.getProperty("out.case",taskProperties.getProperty("out.case",ddlProperties.getProperty("ddl.case","sensitive"))).trim().toLowerCase());


	}
	private ColumnAnalyser newColumnAnalyser() {
		return new ColumnAnalyser(csvFormat.getQualifiedColumnFormats());
	}
	
	private CSVDefinition checkCSVDefinitions() throws IOException {
		List<CSVDefinition> csvDefList=csvFormat.createAllCSVDefinitions();
		LOG.fine("Now Testing "+csvDefList.size()+" csv Definitions ...");
		CSVDefinition[] csvDefinitions=csvDefList.toArray(new CSVDefinition[csvDefList.size()]);
		try (LineNumberReader br = new LineNumberReader(new InputStreamReader(new FileInputStream(inputFile),"UTF-8"))) {
			for(String line=br.readLine();line!=null;line=br.readLine()) {
				if (br.getLineNumber()%100 == 0) {
					csvDefList=new ArrayList<>(csvDefinitions.length);
					for(CSVDefinition csvDefinition : csvDefinitions) 
						if(csvDefinition!=null && !csvDefinition.isRemovable(false))
							csvDefList.add(csvDefinition);
					csvDefinitions=csvDefList.toArray(new CSVDefinition[csvDefList.size()]);
				}
				for(CSVDefinition csvDefinition: csvDefinitions) 
					if(csvDefinition!=null)
						csvDefinition.analyse(br.getLineNumber(), line);
			}
		}
		csvDefList=new ArrayList<>(csvDefinitions.length);
		for(CSVDefinition csvDefinition : csvDefinitions) 
			if(csvDefinition!=null && !csvDefinition.isRemovable(true))
				csvDefList.add(csvDefinition);
		csvDefinitions=csvDefList.toArray(new CSVDefinition[csvDefList.size()]);
		Arrays.sort(csvDefinitions);

		LOG.fine("File might be one of the following csv Formats:");
		for(CSVDefinition csvDefinition : csvDefinitions) 
			LOG.fine("  "+ csvDefinition);
		LOG.fine("Using the 1st for column format analysis.");
		return csvDefinitions[0];
	}
	@Override
	public void run() throws SQLException, IOException {
		LOG.fine(csvFormat.getQualifiedColumnFormats().size()+" FormatDefinitions loaded");
		
		CSVDefinition csvDefinition=checkCSVDefinitions();
		List<ColumnAnalyser> firstColumns=new ArrayList<>(csvDefinition.getNumFirstLineColumns());
		List<ColumnAnalyser> nextColumns=new ArrayList<>(csvDefinition.getNumFirstLineColumns());
		List<String> values;
		List<ColumnAnalyser> columns=firstColumns;
		try (LineNumberReader br = new LineNumberReader(new InputStreamReader(new FileInputStream(inputFile), StandardCharsets.UTF_8))) {
				for(String line=br.readLine();line!=null;line=br.readLine()) {
					try {
						values=csvDefinition.getValues(line);
						if(values!= null) {
							for(int i=0;i<values.size();i++) {
								if(columns.size()<=i)
									columns.add(newColumnAnalyser());
								columns.get(i).analyse(values.get(i));
							}
						}
					} catch(Exception e) {
						// Ignore
					}
					if(br.getLineNumber()==1)
						columns=nextColumns;
				}
		}
		Boolean header=null;
		if(nextColumns.size()>firstColumns.size())
			header=false;
		else 
			for (int i=0;header==null && i<nextColumns.size();i++) 
				header=nextColumns.get(i).isHeader(firstColumns.get(i));
		
		LOG.fine("First line contains column names: "+((header==null)?"maybe":header));
		
		String ddlColumnName;
		String csvColumnName;
		String csvColumnPosition;
		String ddlColumnPosition;
		String tablename = System.getProperty("out.table",taskProperties.getProperty("out.table",taskname));
		try (PrintStream out=new PrintStream(propertiesOutputFile, "ISO-8859-1")) {
			
			out.println("## Part 1 - File Definition");

			out.println("## Part 1.a) Format");
			csvDefinition.write(out, "");
			
			out.println("## csv.header=[false|true]");
			out.println("## if true, the first line of the file contains columnnames otherwise the first line contains data");
			if(Boolean.FALSE.equals(header)) {
				out.println("csv.header=false");
				out.println("");
				out.println("## Part 1.b) Columns");
				out.println("");
				out.println("## columns=[position|name]");
				out.println("## indicates if column format definitions and the mapping to the respective table columns is based on the column's position");
				out.println("## in the file or the columns name. If header=false, columns=name requires a position to name mapping.");
				out.println("columns=position");
			} else {
				out.println("header=true");
				out.println("");
				out.println("## Part 1.b) Columns");
				out.println("");
				out.println("## columns=[position|name]");
				out.println("## indicates if column format definitions and the mapping to the respective table columns is based on the column's position");
				out.println("## in the file or the columns name. If header=false, columns=name requires a position-to-name-mapping (see below).");
				out.println("columns=name");
				out.println("");
				out.println("## columns.(columnname).format=(formatname)");
				out.println("## Define every column's most likely format by it's FormatDefinition Name");
				
				for (int i=0;i<nextColumns.size();i++) {
					csvColumnName=firstColumns.get(i).getHeaderValue();
					nextColumns.get(i).write(out, "column."+csvColumnName+".");
				}
			}
			String prefix ="";
			if(!Boolean.FALSE.equals(header)) {
				prefix="## ";
				out.println("");
				out.println("## In case of columns=position this is defined as ");
			} 

			out.println("");
			out.println("## column.(columnposition).format=(formatname)");
			out.println("## Define every column's most likely format by it's FormatDefinition Name");
			out.println("");
			for (int i=0;i<nextColumns.size();i++) {
				csvColumnPosition=String.valueOf(i+1);
				nextColumns.get(i).write(out, prefix+"column."+csvColumnPosition+".");
			}
			if(!Boolean.FALSE.equals(header))
				prefix="## ";
			out.println("");
			out.println("## column.(columnposition).name=(columnname)");
			out.println("## Define or override a column's name (position-to-name-mapping).");
			out.println("## This is required in case of header=false and columns=name. It can also be used to adjust a columnname.");
			out.println("");
			for (int i=0;i<nextColumns.size();i++) {
				csvColumnPosition=String.valueOf(i+1);
				if (header==null || header)
					csvColumnName=firstColumns.get(i).getHeaderValue();
				else
					csvColumnName="<ColumnName"+csvColumnPosition+">";
				out.println("## column."+csvColumnPosition+".name="+csvColumnName);
			}
			
			out.println("");
			out.println("");
			out.println("## Part 2 - Mapping Definition");

			out.println("");
			out.println("## out.case=[sensitive|upper]");
			out.println("## Defines case sensitivity of the column names.");

			out.println("out.case="+casesensitivity.name());

			out.println("");
			out.println("## out.table=(tablename)");
			out.println("## Defines the name of the target table.");
			out.println("out.table="+tablename);

			out.println("");
			out.println("## columns.map-auto=[true|false]");
			out.println("## Automatic mapping based on column names.");
			out.println("columns.map-auto=false");

			if (header==null || header) {
				out.println("");
				out.println("## map.(tablecolumn)=(columnname)");
				out.println("## maps a column from the the file to a column of the target table.");
				out.println("## (Overrides Automapping.)");
				for (int i=0;i<nextColumns.size();i++) {
					csvColumnName=firstColumns.get(i).getHeaderValue();
					ddlColumnName=casesensitivity.value(csvColumnName);
					out.println("map."+ddlColumnName+"="+csvColumnName);
				}
				out.println("");
				out.println("## In case of columns=position this is defined as ");
				out.println("## map.(tablecolumn)=(columnposition)");
				for (int i=0;i<nextColumns.size();i++) {
					csvColumnPosition=String.valueOf(i+1);
					csvColumnName=firstColumns.get(i).getHeaderValue();
					ddlColumnName=casesensitivity.value(csvColumnName);
					out.println("## map."+ddlColumnName+"="+csvColumnPosition);
				}
			}
			else {
				out.println("");
				out.println("## map.(tablecolumn)=(columnposition)");
				out.println("## maps a column from the the file to a column of the target table.");
				for (int i=0;i<nextColumns.size();i++) {
						
					csvColumnPosition=String.valueOf(i+1);
					ddlColumnPosition=String.valueOf(i+1);

					ddlColumnName=Pattern.compile("\\{position\\}").matcher(ddlProperties.getProperty("ddl.columnname","SPALTE{position}")).replaceAll(ddlColumnPosition);
					out.println("map."+ddlColumnName+"="+csvColumnPosition);
				}
				out.println("");
				out.println("## In case of columns=name this is defined as ");
				out.println("## map.(tablecolumn)=(columnname)");
				for (int i=0;i<nextColumns.size();i++) {
					csvColumnPosition=String.valueOf(i+1);
					ddlColumnPosition=String.valueOf(i+1);
					csvColumnName="<ColumnName"+csvColumnPosition+">";
						
					ddlColumnName=Pattern.compile("\\{position\\}").matcher(ddlProperties.getProperty("ddl.columnname","SPALTE{position}")).replaceAll(ddlColumnPosition);
					out.println("## map."+ddlColumnName+"="+csvColumnName);
				}

			}
			out.flush();
		}

		try (PrintStream out=new PrintStream(ddlOutputFile, "UTF-8")) {
			String ddlColumnFormat;
			String ddlColumn;
			StringBuilder ddlColumns=new StringBuilder();
			
			String ddl=ddlProperties.getProperty(casesensitivity.ddl_create(),"create table {tablename} ({columns})");
			
			for (int i=0;i<nextColumns.size();i++) {
				if (header==null || header) 
					ddlColumnName=casesensitivity.value(firstColumns.get(i).getHeaderValue());
				else
					ddlColumnName=Pattern.compile("\\{position\\}").matcher(ddlProperties.getProperty("ddl.columnname","SPALTE{position}")).replaceAll(String.valueOf(i+1));
				
				ddlColumnFormat=nextColumns.get(i).ddlFormat(ddlProperties);
				ddlColumn = ddlProperties.getProperty(casesensitivity.ddl_column(),"{columnname} {format}");
				ddlColumn = Pattern.compile("\\{columnname\\}").matcher(ddlColumn).replaceAll(ddlColumnName);
				ddlColumn = Pattern.compile("\\{format\\}").matcher(ddlColumn).replaceAll(ddlColumnFormat);

				if(i>0)
					ddlColumns.append(ddlProperties.getProperty("ddl.column.separator",", "));
				ddlColumns.append(ddlColumn);
			}
			ddl= Pattern.compile("\\{tablename\\}").matcher(ddl).replaceAll(tablename);
			ddl= Pattern.compile("\\{columns\\}").matcher(ddl).replaceAll(ddlColumns.toString());
			
			out.println(ddl);

			out.flush();
		}
		
	}
	public static void main(String[] args) {
		/*
		System.setProperty("logging.level", "FINE");
		System.setProperty("logging.simple", "true");
		String[] xargs=(args==null || args.length == 0)?new String[]{"Test1","test","src/examples/data/csv2sql.Person.Mueller.csv"}:args;
		*/
		main(CSVAnalyser.class, args);
	}
	
	/*
	public static void main (String[] args) {
		long START=System.currentTimeMillis();
		Pattern p=Pattern.compile("([^,]*)");
		Pattern s=Pattern.compile(",");
		
		Matcher m=p.matcher("1,2,3,,,6,,");
		while (m.lookingAt()) {
			System.out.println("P=<"+m.group(1)+">");
			if(m.end()==m.regionEnd())
				break;
			m.region(m.end(), m.regionEnd());
			m.usePattern(s);
			if(!m.lookingAt())
				throw new RuntimeException("Fehler");
			m.region(m.end(), m.regionEnd());
			m.usePattern(p);
		}
		long END=System.currentTimeMillis();
		System.out.println((END-START)+" millis");
	}
*/
	/*
	public static void main (String[] args) {
		CSVDefinition csv=new CSVDefinition(new SeparatorDefinition(null, "abc"), new EscapeDefinition(null, "$"), new BoundaryDefinition(null, "((", "))"));
		
		//Pattern p=Pattern.compile("(.*(?!abc))");
		long START=System.currentTimeMillis();
		Pattern s=Pattern.compile("abc");
		Pattern l=Pattern.compile("\\(\\(");
		Pattern r=Pattern.compile("\\)\\)");
		Pattern e=Pattern.compile("(^|[^\\$])\\$(\\$\\$)*[^\\$]");
		
//		String val="1abc((2))abc((3$$abc$)))abcabc5abcabc";
		String val="((3$$abc$)))abcabc5abcabc((x$))";
		Matcher m=l.matcher(val);
		Matcher em=e.matcher(val);
		
		int valStart=0;
		int valEnd=0;
		int escaped=Integer.MAX_VALUE;
		if(em.find()) {
			escaped=em.end()-1;
			em.region(escaped+1, em.regionEnd());
		} 
			
		boolean sepFound=false;
		int endStart=0;
		for (int i=1;;i++) {
			m.usePattern(l);
			if(m.lookingAt()) {
				valStart=m.end();
				m.region(m.end(), m.regionEnd());
				m.usePattern(r);
				for(boolean endFound=false;!endFound;) {
					endFound=m.find();
					if(!endFound)
						throw new RuntimeException("Position "+ m.regionStart() +": Closing ColumnBoundaries "+r.pattern()+" not found.");
					endStart=m.start();
					while (escaped<endStart) {
						if(em.find()) {
							escaped=em.end()-1;
							em.region(escaped+1, em.regionEnd());
						} else 
							escaped=Integer.MAX_VALUE;
					}
					if (escaped==endStart) {
						endFound=false;
						m.region(escaped+1, m.regionEnd());
					}
				}
				valEnd=endStart;
				m.region(m.end(), m.regionEnd());
				
				m.usePattern(s);
				sepFound=m.lookingAt();
				if(!sepFound && m.regionStart() < m.regionEnd() )
					throw new RuntimeException("Position "+ m.regionStart() +": EOL or ColumnSeparator "+s.pattern()+" expected.");
			} else {
				valStart=m.regionStart();
				m.usePattern(s);
				for(boolean endFound=false;!endFound;) {
					sepFound=m.find();
					if(!sepFound)
						break;
					endStart=m.start();
					while (escaped<endStart) {
						if(em.find()) {
							escaped=em.end()-1;
							em.region(escaped+1, em.regionEnd());
						} else 
							escaped=Integer.MAX_VALUE;
					}
					endFound = (escaped != endStart);
					if(!endFound)
						m.region(escaped+1, m.regionEnd());
				}
				valEnd=sepFound?m.start():m.regionEnd();
			}
			System.out.println(i+": <"+val.substring(valStart, valEnd)+">");
			if(!sepFound)
				break;
			m.region(m.end(), m.regionEnd());
		}
		long END=System.currentTimeMillis();
		System.out.println((END-START)+" millis");

	}
	*/
	/*
	public static void main (String[] args) {
		long START=System.currentTimeMillis();
		CSVDefinition csv=new CSVDefinition(new SeparatorDefinition(null, "abc"), new EscapeDefinition(null, "$"), new QuotesDefinition(null, "((", "))"));
//		String val="1abc((2))abc((3$$abc$)))abcabc5abcabc";
		String val="((3$$$$$abc$)))abcabc5abcabc((x))";

		List<String> cols=csv.getValues(val);
		for (int i=0;i<cols.size();i++)
			System.out.println(i+": <"+cols.get(i)+">");
		
		long END=System.currentTimeMillis();
		System.out.println((END-START)+" millis");

	}
	*/
}
