/*
 * Copyright 2023, Rainer Noack, Frankfurt am Main, Germany.
 *
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. *
 */
package com.enitsys.sqltools;

import java.io.Reader;
import java.io.StringReader;
import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;

final class DBMSOutput implements AutoCloseable {
    private Connection connection;
    private final boolean isOracle;
    DBMSOutput(Connection connection, int size) throws SQLException {
        this.connection = connection;
        this.isOracle = "Oracle".equals(connection.getMetaData().getDatabaseProductName());
        if (isOracle)
        	try (Statement stmt = connection.createStatement()) {
        		stmt.execute("begin dbms_output.enable(" + size + "); end;");
        	}
    }
    Reader fetch() throws SQLException {
        if (connection==null)
            throw new RuntimeException("DBMSOutput already closed");
        if (!isOracle)
        	return new StringReader("");
        try (CallableStatement stmt = connection.prepareCall(
                "declare l_line varchar2(255); " +
                        " l_done number; " +
                        " l_buffer long; " +
                        "begin " +
                        " loop " +
                        " exit when length(l_buffer)+255 > :maxbytes OR l_done = 1; " +
                        " dbms_output.get_line( l_line, l_done ); " +
                        " l_buffer := l_buffer || l_line || chr(10); " +
                        " end loop; " +
                        " :done := l_done; " +
                        " :buffer := l_buffer; " +
                        "end;");) {

            stmt.registerOutParameter(2, java.sql.Types.INTEGER);
            stmt.registerOutParameter(3, java.sql.Types.VARCHAR);
            StringBuilder stringBuilder = new StringBuilder(32000);
            String string;
            int done = 0;
            while (done == 0) {
                stmt.setInt(1, 32000);
                stmt.executeUpdate();
                done = stmt.getInt(2);
                string = stmt.getString(3);
                if (string != null && !string.isEmpty())
                    stringBuilder.append(string);
            }
            return new StringReader(stringBuilder.toString());
        }
    }

    Reader fetchAndClose() throws SQLException {
        Reader result = fetch();
        close();
        return result;
    }


    /*
     * close closes the callable statements associated with
     * the DbmsOutput class. Call this if you allocate a DbmsOutput
     * statement on the stack and it is going to go out of scope --
     * just as you would with any callable statement, result set
     * and so on.
     */
    public void close() throws SQLException {
        if (connection!=null) {
            try {
            	if (isOracle)
            		try (Statement stmt = connection.createStatement()) {
            			stmt.execute("begin dbms_output.disable; end;");
            		}
            } finally {
                connection = null;
            }
        }
    }

    @Override
    protected void finalize() throws Throwable {
        super.finalize();
        try {
            close();
        } catch (Exception e) {
            //ignore
        }
    }
}

