/*
 * Copyright 2023, Rainer Noack, Frankfurt am Main, Germany.
 *
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. *
 */
package com.enitsys.sqltools;

import java.io.IOException;
import java.io.OutputStream;
import java.util.logging.Level;
import java.util.logging.Logger;

final class LoggerOutputStream extends OutputStream {
    private static final int DEFAULT_BUFFER_LENGTH = 2048;
    private static final String LINE_SEPARATOR = System.getProperty("line.separator");
    private final Logger log;
    private final Level level;
    private boolean closed = false;
    private byte[] buf;
    private int count;
    /**
     * Remembers the size of the buffer for speed.
     */
    private int bufLength;

    LoggerOutputStream(Logger log, Level level) {
        if (log == null)
            throw new NullPointerException("log");

        this.level = level;
        this.log = log;
        bufLength = DEFAULT_BUFFER_LENGTH;
        buf = new byte[DEFAULT_BUFFER_LENGTH];
        count = 0;
    }
    @Override
    public void close() {
        if (closed)
            return;
        flush();
        closed = true;
    }
    @Override
    public void write(final int b) throws IOException {
        if (closed)
            throw new IOException("The stream has been closed.");

        // don't log nulls
        if (b == 0)
            return;

        // would this be writing past the buffer?
        if (count == bufLength) {
            // grow the buffer
            final int newBufLength = bufLength + DEFAULT_BUFFER_LENGTH;
            final byte[] newBuf = new byte[newBufLength];

            System.arraycopy(buf, 0, newBuf, 0, bufLength);

            buf = newBuf;
            bufLength = newBufLength;
        }

        buf[count] = (byte) b;
        count++;
    }

    /**
     * Flushes this output stream and forces any buffered output bytes to be
     * written out. The general contract of <code>flush</code> is that
     * calling it is an indication that, if any bytes previously written
     * have been buffered by the implementation of the output stream, such
     * bytes should immediately be written to their intended destination.
     */
    @Override
    public void flush() {
        if (count == 0)
            return;

        // suppress blank lines; flushing from PrintStream puts out these
        if (count == LINE_SEPARATOR.length()) {
            if (((char) buf[0]) == LINE_SEPARATOR.charAt(0)
                    && (count == 1 || ((char) buf[1]) == LINE_SEPARATOR.charAt(1))) {
                reset();
                return;
            }
        }
        final byte[] theBytes = new byte[count];
        System.arraycopy(buf, 0, theBytes, 0, count);
        log.log(level,new String(theBytes));
        reset();
    }
    private void reset() {
        count = 0;
    }
    @Override
    protected void finalize() throws Throwable {
        close();
        super.finalize();
    }
}