/*
 * Copyright 2023, Rainer Noack, Frankfurt am Main, Germany.
 *
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. *
 */
package com.enitsys.sqltools;

import java.io.*;
import java.nio.file.Files;
import java.sql.*;
import java.util.zip.ZipOutputStream;

public final class SQL2file extends AbstractSQLTool implements Archivable {
    private static final int contenttype_varchar = 1;
    private static final int contenttype_clob = 2;

    private final String stmtString;

    SQL2file(String taskname, String connection, String directory) throws IOException {
        super(taskname, connection, directory);
        this.stmtString = stmt(null, taskname + ".sql");
    }
   protected void encode(ResultSet rs, int colContent, int contentType, PrintWriter out) throws SQLException, IOException {
            if (contentType == contenttype_varchar) {
                String s = rs.getString(colContent);
                if (s != null)
                    out.println(s);
            } else if (contentType == contenttype_clob) {
                Clob clob = rs.getClob(colContent);
                if (clob != null) {
                    BufferedReader r = new BufferedReader(clob.getCharacterStream());
                    for (String s = r.readLine(); s != null; s = r.readLine())
                        out.println(s);
                }
            } else {
                Object o = rs.getObject(colContent);
                if (o != null)
                    out.println(o.toString());
            }
            out.flush();
    }


    public void write(Connection dataConnection, OutputStreamFactory outputStreamFactory, boolean addSQLFile) throws SQLException, IOException {
        if (dataConnection == null) {
            try (Connection conn =createConnection()) {
                write(conn, outputStreamFactory, addSQLFile);
            }
            return;
        }
        if (addSQLFile) {
            try (PrintWriter printWriter = new ModifiablePrintWriter(new OutputStreamWriter(outputStreamFactory.newOutputStream(taskname + ".sql"), taskProperties.getProperty("out.encoding", "UTF-8")), null)) {
                printWriter.println(stmtString);
                printWriter.flush();
            }
        }
        try (Statement stmt = dataConnection.createStatement()) {
            String fetchsize=taskProperties.getProperty("sql.fetchsize");
            if (fetchsize!=null)
                stmt.setFetchSize(Integer.parseInt(fetchsize));
            try (ResultSet rs = stmt.executeQuery(stmtString)) {
                ResultSetMetaData md = rs.getMetaData();
                final int colcount = md.getColumnCount();
                if (colcount < 2 )
                    throw new RuntimeException("SQL statement returns less then 2 columns.");
                int contenttype = contenttype_varchar;
                if (md.getColumnType(2) == Types.CLOB) {
                    contenttype = contenttype_clob;
                    LOG.fine("Running in CLOB mode.");
                }
                while(rs.next()) {
                    String filename=rs.getString(1);
                    System.out.print("Writing file "+outputStreamFactory.getFilename(filename)+" ...");
                    try (PrintWriter printWriter = new ModifiablePrintWriter(new OutputStreamWriter(outputStreamFactory.newOutputStream(filename), taskProperties.getProperty("out.encoding", "UTF-8")), null)) {
                        encode(rs, 2, contenttype, printWriter);
                    }
                    System.out.println(" done.");
                }
            }
        }

    }
    public void run() throws SQLException, IOException {
        try (Connection conn = createConnection()) {
            String zipFile = replace(System.getProperty("out.zip.file", taskProperties.getProperty("out.zip.file")), replacements);
            if (zipFile != null) {
                File outfile = (destDir == null || new File(zipFile).isAbsolute()) ? new File(zipFile) : new File(destDir, zipFile);
                Files.createDirectories(outfile.toPath().getParent());
                try (OutputStreamFactory outputStreamFactory = new OutputStreamFactory.ZipEntryOutputStreamFactory(new ZipOutputStream(new FileOutputStream(outfile)))) {
                    write(conn, outputStreamFactory, false);
                }
            } else {
                String tempDir = System.getProperty("out.tempdir", taskProperties.getProperty("out.tempdir"));
                OutputStreamFactory outputStreamFactory = null;
                if (tempDir != null)
                    outputStreamFactory = new OutputStreamFactory.TempOutputStreamFactory(destDir, tempDir);
                else
                    outputStreamFactory = new OutputStreamFactory.DirOutputStreamFactory(destDir);
                write(conn, outputStreamFactory, false);
            }
        }
    }

    public static void main(String[] args) {
        main(SQL2file.class, args);
    }

}
